require "defines"
require "config"

--[[
    This module handles the gameloop alteration, detecting fish entities wandering around or in front of fishing poles, and replacing them with fish items.

]]--

--[[
    
]]--

function removeKeys(myTable,toRemove)
    local n=#myTable

    for x=1,n do
        if toRemove[x] then
            myTable[x]=nil
            --globalPrint("["..x.."] = nil")
        end
    end

    local y=0
    for x=1,n do
        if myTable[x]~=nil then
            y=y+1
            myTable[y]=myTable[x]
            --globalPrint("["..y.."] = ["..x.."]")
        end
    end
    for x=y+1,n do
        myTable[x]=nil
        --globalPrint("["..x.."] = nil")
    end
    return myTable
end

function addPos(p1,p2)
  if not p1.x then
    error("Invalid position", 2)
  end
  if p2 and not p2.x then
    error("Invalid position 2", 2)
  end
  local p2 = p2 or {x=0,y=0}
  return {x=p1.x+p2.x, y=p1.y+p2.y}
end
function subPos(p1,p2)
  if not p1.x then
    error("Invalid position", 2)
  end
  if p2 and not p2.x then
    error("Invalid position 2", 2)
  end
  local p2 = p2 or {x=0,y=0}
  return {x=p1.x-p2.x, y=p1.y-p2.y}
end

function verifySettings()
	if (RanaMods.Fishing_Pole.config.pole.tickRate < 0) then
		RanaMods.Fishing_Pole.config.pole.tickRate = 0
		throwError("Tick rate must be >= 0.")
	end
end

function globalPrint(msg)
  local players = game.players
  for x=1, #players do
    players[x].print(msg)
  end
end

script.on_init(onload)
script.on_load(onload)

function onload()
	if (global.loaded == nil) then
		global.loaded = true
		verifySettings()
	end
end

function onCreate(event)
    local pole = event.created_entity
    
    if pole.name == "fishing-pole" then
        if global.pole == nil then
            global.pole = {}
        end
        table.insert(global.pole, pole)
    end
end

function tick()
    local toRemove = {}
    
	--[[if (game.tick % RanaMods.Fishing_Pole.config.spawn.tickRate) == 0 then
            local fishCount = 0
            for _, entity in ipairs(game.surfaces.nauvis.find_entities_filtered
                {
                    area =
                    {
                        {-1028 ,-1028 },
                        {1028 ,1028 }
                    },
                    name="fish"
                }
            )do
                fishCount = fishCount+1
            end
            globalPrint("FishCount "..fishCount)]]--
        --[[for z, surface in pairs(game.surfaces) do
            for coord in surface.get_chunks() do
                local X,Y = coord.x, coord.y
                if surface.is_chunk_generated{X,Y} then
                    local area = {{X*32, Y*32}, {X*32 + 32, Y*32 + 32}}
                    -- Quick Method. Select random tile If deep water or Shallow water, 1/60 chance of fish
                    local selectedTile = {math.random(area[1][1],area[2][1]),math.random(area[1][2],area[2][2])}
                    local chance = 1.0
                    
                    local tileName = surface.get_tile(selectedTile[1],selectedTile[2]).name
                    if tileName == "deepwater" then
                        chance = chance
                    elseif tileName == "water" then
                        chance = chance-RanaMods.Fishing_Pole.config.spawn.chanceShallowPenalty
                    else
                        chance = 0.00 -- If not water, Never spawn fishies.
                    end
                    
                    chance = chance - surface.get_pollution({X,Y})*RanaMods.Fishing_Pole.config.spawn.chancePollutionPenalty
                    if chance > 0 then
                        chance = chance*RanaMods.Fishing_Pole.config.spawn.chanceBase
                        local rand = math.random()
                        if rand <= chance then
                            surface.create_entity{name = "fish", position = selectedTile, force=game.forces.neutral}
                            globalPrint("Spawned Fish("..selectedTile[1]..","..selectedTile[2]..")"..chance..","..rand)
                            
                            
                        end
                    end
                end
            end
        end]]--
   -- end
	if (game.tick % RanaMods.Fishing_Pole.config.pole.tickRate) == 0 then
        --globalPrint("tick:"..game.tick)
        if global.pole ~= nil then
            local array = {}
            --globalPrint("hasPole")
            for k,pole in pairs(global.pole) do
                if pole.valid then 
                    
                    
                    local surface = pole.surface
                    local pos = pole.position
                    pos = addPos(pos,RanaMods.Fishing_Pole.config.pole.offset[pole.direction])
                    
                    --globalPrint("energy = "..pole.energy)
                    if pole.energy > 0 then -- Needs better energy gauge. ~350.something is cap, why? value in prototype is 0.4kw.
                    
                        --if pole.held_stack == nil then -- Invalid check, figure out why
                        
                            local tileName = surface.get_tile(pos.x, pos.y).name
                            
                            if ((tileName == "water") or (tileName == "deepwater")) then -- validate target location
                                --[[local existingItem
                                for _, entity in ipairs(surface.find_entities_filtered
                                    {
                                        area =
                                        {
                                            {pos.x-32,pos.y-32},
                                            {pos.x+32,pos.y+32}
                                        },
                                        name="item-on-ground"
                                    }
                                )do
                                    
                                end]]--
                                if true then -- Validate no existing fish item
                                    -- Validate fish in range
                                    -- For each, break on first.
                                    local fish
                                    for _, entity in ipairs(surface.find_entities_filtered
                                        {
                                            area =
                                            {
                                                {pos.x-RanaMods.Fishing_Pole.config.pole.pullRadius ,pos.y-RanaMods.Fishing_Pole.config.pole.pullRadius },
                                                {pos.x+RanaMods.Fishing_Pole.config.pole.pullRadius ,pos.y+RanaMods.Fishing_Pole.config.pole.pullRadius }
                                            },
                                            name="fish"
                                        }
                                    )do
                                        fish = entity
                                        --globalPrint("fish")
                                        break
                                    end
                                    
                                    
                                    if fish ~= nil then
                                        local fishDifference = subPos(fish.position,pos)
                                        --globalPrint("pos ("..pos.x..","..pos.y..")")
                                        --globalPrint("fish ("..fish.position.x..","..fish.position.y..")")
                                        --globalPrint("Dif ("..fishDifference.x..","..fishDifference.y..")")
                                        if (math.abs(fishDifference.x) < RanaMods.Fishing_Pole.config.pole.pickupRadius) and (math.abs(fishDifference.y) < RanaMods.Fishing_Pole.config.pole.pickupRadius) then
                                            --globalPrint("fish+")
                                            fish.destroy()
                                            surface.create_entity{name = "item-on-ground", position=pos, stack = {name="raw-fish"}}
                                        else
                                            local newPos = fish.position
                                            if fishDifference.x > 0 then
                                                newPos.x = math.max(fish.position.x-RanaMods.Fishing_Pole.config.pole.impulse,pos.x)
                                            else
                                                newPos.x = math.min(fish.position.x+RanaMods.Fishing_Pole.config.pole.impulse,pos.x)
                                            end
                                            
                                            if fishDifference.y > 0 then
                                                newPos.y = math.max(fish.position.y-RanaMods.Fishing_Pole.config.pole.impulse,pos.y)
                                            else
                                                newPos.y = math.min(fish.position.y+RanaMods.Fishing_Pole.config.pole.impulse,pos.y)
                                            end
                                            fish.teleport(newPos)
                                        end
                                    else
                                        --globalPrint("NoFish")
                                        local x = math.floor(pos.x/32)
                                        if (x < 0) then
                                            x = (-x*2) + 1
                                        
                                        else
                                            x = x*2
                                            
                                        end
                                        
                                        local y = math.floor(pos.y/32)
                                        if (y < 0) then
                                            y = (-y*2) + 1
                                        
                                        else
                                            y = y*2
                                            
                                        end
                                        
                                        if array[x] then
                                            
                                        else
                                            array[x] = {}
                                        end
                                        if (array[x][y]) then
                                            --globalPrint("false "..x.." "..y.." ("..pos.x..","..pos.y..")")
                                        else
                                            --globalPrint("true "..x.." "..y.." ("..pos.x..","..pos.y..")")
                                            array[x][y] = true
                                        end
                                        
                                        -- No Fish :(
                                    end
                                else
                                    -- No existing fish item
                                end
                            else
                                -- Not at water
                            end
                        --else
                            -- Has fish or otherwise already
                        --end
                    else
                        
                        -- Not enough energy
                    end
                else
                    -- Doesn't exist, flag for removal.
                    table.insert(toRemove, k)
                    --globalPrint("Removed "..k)
                end
            end
        end
        if next(toRemove) then
                    --globalPrint("Removed ~")
            global.pole = removeKeys(global.pole,toRemove)
        end
    end
end

script.on_event(defines.events.on_built_entity, onCreate)
script.on_event(defines.events.on_robot_built_entity, onCreate)

script.on_event(defines.events.on_tick,tick)